from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Tuple, cast, no_type_check

import pytest

from tierkreis.builder import (
    Break,
    Case,
    Const,
    Continue,
    Copyable,
    Else,
    If,
    IfElse,
    Input,
    MakeTuple,
    Match,
    Namespace,
    Output,
    Scope,
    Tag,
    UnpackTuple,
    ValueSource,
    closure,
    current_builder,
    current_graph,
    graph,
    lazy_graph,
    loop,
)
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core import Labels
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    GraphValue,
    TierkreisGraph,
)
from tierkreis.core.types import (
    IntType,
    MapType,
    StringType,
    TierkreisPair,
    UnpackRow,
    VecType,
)
from tierkreis.core.utils import map_vals, rename_ports_graph
from tierkreis.core.values import (
    BoolValue,
    FloatValue,
    IntValue,
    MapValue,
    PairValue,
    StringValue,
    StructValue,
    TierkreisValue,
    TierkreisVariant,
    VariantValue,
    VecValue,
)
from tierkreis.pyruntime import PyRuntime

if TYPE_CHECKING:
    from tierkreis.builder import StablePortFunc
    from tierkreis.core.tierkreis_graph import NodePort, NodeRef


def _compare_graphs(
    first: TierkreisGraph,
    second: TierkreisGraph,
) -> bool:
    return first.to_proto() == second.to_proto()


def _vecs_graph() -> TierkreisGraph:
    tg = TierkreisGraph("g")
    con = tg.add_const([2, 4])
    tg.set_outputs(value=con)
    tg.annotate_output(Labels.VALUE, VecType(IntType()))
    tg.output_order = [Labels.VALUE]
    return tg


@pytest.fixture()
def _vecs_graph_builder() -> TierkreisGraph:
    @graph()
    def g() -> Output[VecValue[IntValue]]:
        return Output(Const([2, 4]))

    return g


def _structs_graph() -> TierkreisGraph:
    tg = TierkreisGraph("g")
    factory = tg.add_func(
        "make_struct",
        **map_vals(dict(height=12.3, name="hello", age=23), tg.add_const),
    )
    sturct = tg.add_func("unpack_struct", struct=factory["struct"])

    tg.set_outputs(value=sturct["age"])
    tg.annotate_output("value", IntType())
    tg.output_order = [Labels.VALUE]

    return tg


@pytest.fixture()
def _structs_graph_builder(bi) -> TierkreisGraph:
    @graph()
    def g() -> Output[IntValue]:
        st_node = bi.make_struct(height=Const(12.3), name=Const("hello"), age=Const(23))
        return Output(st_node["struct"]["age"])

    return g


def _maps_graph() -> TierkreisGraph:
    tg = TierkreisGraph("g")
    mp_val = tg.add_func("remove_key", map=tg.input["mp"], key=tg.add_const(3))
    ins = tg.add_func(
        "insert_key",
        map=mp_val["map"],
        **map_vals(dict(key=5, val="bar"), tg.add_const),
    )

    tg.set_outputs(mp=ins["map"], vl=mp_val["val"])
    tg.annotate_input("mp", MapType(IntType(), StringType()))
    tg.annotate_output("mp", MapType(IntType(), StringType()))
    tg.annotate_output("vl", StringType())
    tg.input_order = ["mp"]
    tg.output_order = ["mp", "vl"]
    return tg


@pytest.fixture()
def _maps_graph_builder(bi) -> TierkreisGraph:
    @dataclass
    class Mapout(UnpackRow):
        mp: MapValue[IntValue, StringValue]
        vl: StringValue

    @graph()
    def g(mp: Input[MapValue[IntValue, StringValue]]) -> Output[Mapout]:
        mp, val = bi.remove_key(mp, Const(3))
        return Output(bi.insert_key(mp, Const(5), Const("bar")), val)

    return g


def _tag_match_graph() -> TierkreisGraph:
    id_graph = TierkreisGraph("foo")
    id_graph.set_outputs(value=id_graph.input[Labels.VALUE])

    tg = TierkreisGraph("g")
    in_v = tg.add_tag("foo", value=tg.add_const(4))
    m = tg.add_match(in_v, foo=tg.add_const(id_graph))
    e = tg.add_func("eval", thunk=m[Labels.THUNK])
    tg.set_outputs(value=e[Labels.VALUE])
    tg.annotate_output("value", IntType())
    tg.output_order = [Labels.VALUE]

    return tg


@pytest.fixture()
def _tag_match_graph_builder(bi) -> TierkreisGraph:
    @graph()
    def g() -> Output[IntValue]:
        with Match(Tag("foo", Const(4))) as match:
            with Case("foo") as h1:
                Output(h1.var_value)
        return Output(match.nref)

    return g


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "builder,expected_gen",
    [
        ("_vecs_graph_builder", _vecs_graph),
        ("_structs_graph_builder", _structs_graph),
        ("_maps_graph_builder", _maps_graph),
        ("_tag_match_graph_builder", _tag_match_graph),
    ],
)
async def test_builder_sample(
    builder: str,
    expected_gen: Callable[[], TierkreisGraph],
    client: RuntimeClient,
    bi,  # for some reason fails without this
    request,
) -> None:
    tg = request.getfixturevalue(builder)

    assert _compare_graphs(tg, expected_gen())
    if client.can_type_check:
        tg = await client.type_check_graph(tg)


def _big_sample_builder(bi: Namespace) -> TierkreisGraph:
    def add2(x: ValueSource) -> Output[IntValue]:
        return Output(bi.iadd(x, Const(2)))

    def add5(x: ValueSource) -> Output[IntValue]:
        return Output(bi.iadd(x, Const(5)))

    @dataclass
    class Point:
        p1: FloatValue
        p2: IntValue

    @graph()
    def struc_id(in_st: Input[StructValue[Point]]) -> Output[StructValue[Point]]:
        return Output(in_st)

    @graph()
    def func(v1: Input[IntValue], v2: Input) -> Output:
        fst, scd = UnpackTuple(v2, 2)
        other_total = bi.iadd(fst, Const(3))
        dbl = double(bi)
        _tuple_out = MakeTuple(Const(True), Const("asdf"))
        total = dbl(v1)

        quadruple = bi.sequence(Const(dbl), Const(dbl))

        total4 = bi.eval(quadruple, value=total)

        with IfElse(scd) as ifelse:
            with If():
                add2(total4)
            with Else():
                add5(total4)

        @loop()
        def loop_def(initial_: Input[IntValue]) -> Output:
            initial = Copyable(initial_)
            with IfElse(bi.ilt(initial, Const(100))) as loop_ifelse:
                with If():
                    Continue(bi.iadd(initial, Const(5)))
                with Else():
                    Break(initial)
            return Output(loop_ifelse.nref)

        _disc = struc_id(Const(Point(FloatValue(4.3e1), IntValue(3))))

        return Output(o1=ifelse.nref, o2=loop_def(other_total))

    return func


@pytest.mark.asyncio
async def test_bigexample(client: RuntimeClient, bi) -> None:
    tg = _big_sample_builder(bi)
    if not client.can_type_check:
        pytest.skip()
    tg = await client.type_check_graph(tg)
    assert sum(1 for _ in tg.nodes()) == 24
    for flag in (True, False):
        outputs = await client.run_graph(tg, v1=67, v2=(45, flag))
        outputs = await client.run_graph(tg, v1=67, v2=(45, flag))
        pyouts = {key: val.try_autopython() for key, val in outputs.items()}
        assert pyouts == {"o2": 103, "o1": 536 + (2 if flag else 5)}


def double(bi: Namespace) -> TierkreisGraph:
    @graph()
    def _double(value: Input[IntValue]) -> Output[IntValue]:
        return Output(bi.iadd(*bi.copy(value)))

    return _double


@pytest.mark.asyncio
async def test_double(client: RuntimeClient, bi: Namespace):
    out = await client.run_graph(double(bi), value=10)
    assert out == {"value": IntValue(20)}


@pytest.mark.asyncio
async def test_copy(client: RuntimeClient, bi: Namespace):
    @dataclass
    class CopyOut(UnpackRow):
        a: Any
        b: IntValue

    @graph()
    def copy_graph(y: Input[IntValue]) -> Output[CopyOut]:
        return Output(*bi.copy(y))

    out = await client.run_graph(copy_graph, y=10)
    assert out == {"a": IntValue(10), "b": IntValue(10)}


@pytest.mark.asyncio
async def test_ifelse(client: RuntimeClient, bi: Namespace):
    # some graph type annotations are missing here to test this scenario
    @graph()
    def triple(y: Input[IntValue]):
        return Output(bi.imul(y, Const(3)))

    @graph()
    def ifelse(x: Input, flag) -> Output:
        bias = Const(1)
        with IfElse(flag) as sw:
            with If():
                Output(value=bi.iadd(bias, triple(x)))
            with Else():
                Output(value=x)
        return Output(sw.nref)

    outs = [await client.run_graph(ifelse, x=2, flag=f) for f in (True, False)]

    assert outs[0] == {"value": IntValue(7)}
    assert outs[1] == {"value": IntValue(2)}


def num_copies(g: TierkreisGraph) -> int:
    return sum(1 for n in g.nodes() if n.is_copy_node())


def constant_subgraphs(g: TierkreisGraph) -> list[TierkreisGraph]:
    return [
        n.value.value
        for n in g.nodes()
        if isinstance(n, ConstNode) and isinstance(n.value, GraphValue)
    ]


@pytest.mark.asyncio
async def test_ifelse_copying(
    client: RuntimeClient,
    bi: Namespace,
):
    @graph()
    def ifelse(x: Input[IntValue]) -> Output[IntValue]:
        pred = bi.eq(bi.imod(x, Const(2)), Const(0))
        x2 = Copyable(x)
        with IfElse(pred) as sw:
            with If():
                Output(value=bi.idiv(x2, Const(2)))
            with Else():
                Output(value=bi.iadd(bi.imul(x2, Const(3)), Const(1)))
        return Output(sw.nref)

    assert num_copies(ifelse) == 1
    subgraphs = constant_subgraphs(ifelse)
    assert len(subgraphs) == 2
    for g in subgraphs:
        assert num_copies(g) == 0

    outs = await client.run_graph(ifelse, x=10)
    assert outs == {"value": IntValue(5)}
    outs = await client.run_graph(ifelse, x=5)
    assert outs == {"value": IntValue(16)}


@pytest.mark.asyncio
async def test_copy_twice_inside_if(client: RuntimeClient, bi: Namespace):
    @graph()
    def ifelse(x: Input[IntValue], flag: Input[BoolValue]) -> Output[IntValue]:
        with IfElse(flag) as sw:
            with If():
                c = Copyable(x)  # Copy in inner graph
                Output(value=bi.imul(c, bi.iadd(c, Const(1))))
            with Else():
                Output(value=x)
        return Output(sw.nref)

    assert num_copies(ifelse) == 0
    assert sorted(num_copies(g) for g in constant_subgraphs(ifelse)) == sorted([1, 0])

    outs = {b: await client.run_graph(ifelse, x=3, flag=b) for b in [True, False]}
    assert outs[True] == {"value": IntValue(12)}
    assert outs[False] == {"value": IntValue(3)}


@pytest.mark.asyncio
async def test_copy_twice_from_outside_if(bi: Namespace):
    @lazy_graph()
    def ifelse(x: Input[IntValue], flag: Input[BoolValue]) -> Output[IntValue]:
        c = Copyable(x)  # This can only copy in outer graph
        with IfElse(flag) as sw:
            with If():
                # Use copy twice in inner graph, this is not allowed
                Output(value=bi.imul(c, bi.iadd(c, Const(1))))
            with Else():
                Output(value=c)
        return Output(sw.nref)

    with pytest.raises(ValueError, match="Already captured"):
        _ = ifelse.graph


@pytest.mark.asyncio
async def test_loop(client: RuntimeClient, bi: Namespace):
    @graph()
    def loopyg() -> Output[FloatValue]:
        incr = Const(1)

        @loop()
        def loop_def(value: Input[IntValue]) -> Output:
            x1, x2 = bi.copy(value)

            @closure()
            def succ(x: Input[IntValue]) -> Output[IntValue]:
                return Output(value=bi.iadd(x, incr))

            with IfElse(bi.ilt(x1, Const(7))) as sw:
                with If():
                    Continue(succ(x2))
                with Else():
                    Break(bi.int_to_float(x2))
            return Output(sw.nref)

        return Output(loop_def(Const(0)))

    assert (await client.run_graph(loopyg)) == {"value": FloatValue(7.0)}


@pytest.mark.asyncio
async def test_match(client: RuntimeClient, bi: Namespace):
    @dataclass
    class _Vdict:
        list: VecValue[FloatValue]
        pair: PairValue[FloatValue, FloatValue]

    @graph()
    def match(var: Input[VariantValue[_Vdict]]) -> Output:
        factor = Const(2.0)
        with Match(var) as match:
            with Case("list") as h1:
                lst, fst = bi.pop(h1.var_value)
                Output(bi.push(lst, bi.fmul(fst, factor)))
            with Case("pair") as h2:
                fst, scd = bi.unpack_pair(h2.var_value)
                Output(bi.push(bi.push(Const([]), bi.fmul(fst, factor)), scd))
        return Output(match.nref)

    assert (
        await client.run_graph(
            match, var=TierkreisVariant("pair", TierkreisPair(1.0, 2.0))
        )
    ) == {"value": VecValue(values=[FloatValue(value=2.0), FloatValue(value=2.0)])}
    assert (
        await client.run_graph(match, var=TierkreisVariant("list", [1.0, 2.0]))
    ) == {"value": VecValue(values=[FloatValue(value=1.0), FloatValue(value=4.0)])}


@pytest.fixture()
def _tuple_builder(bi: Namespace, sig: Signature) -> TierkreisGraph:
    @dataclass
    class TestPair(UnpackRow):
        first: IntValue
        second: StringValue

    @graph()
    def main() -> Output[TestPair]:
        return Output(*UnpackTuple(Const((2, "asdf")), 2))

    return main


@pytest.fixture()
def _if_no_inputs(bi: Namespace, sig: Signature) -> TierkreisGraph:
    @graph()
    def main(pred: Input[BoolValue]) -> Output[IntValue]:
        with IfElse(pred) as ifelse:
            with If():
                Output(Const(3))
            with Else():
                Output(Const(5))
        return Output(ifelse.nref)

    return main


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "builder,inputs,expected_outputs",
    [
        ("_tuple_builder", {}, {"first": IntValue(2), "second": StringValue("asdf")}),
        ("_if_no_inputs", {"pred": BoolValue(True)}, {"value": IntValue(3)}),
        ("_if_no_inputs", {"pred": BoolValue(False)}, {"value": IntValue(5)}),
    ],
)
async def test_run_sample(
    client: RuntimeClient,
    builder: str,
    inputs: dict[str, TierkreisValue],
    expected_outputs: dict[str, TierkreisValue],
    bi,
    request,
) -> None:
    tg = request.getfixturevalue(builder)

    outputs = await client.run_graph(tg, **inputs)
    assert outputs == expected_outputs


@pytest.mark.asyncio
async def test_Copyable(bi, client: RuntimeClient) -> None:
    @graph()
    def foo(a: Input[IntValue], b: Input[IntValue]) -> Output:
        assert num_copies(current_graph()) == 0

        # Even without an explicit Copy, discards should be removed
        f = bi.iadd(a=a, b=b)
        assert num_copies(current_graph()) == 0

        a_squared_plus_ab = bi.imul(a=Copyable(a), b=f)
        assert num_copies(current_graph()) == 1

        b_plus_2a = bi.iadd(a=Copyable(a), b=Copyable(f))
        assert num_copies(current_graph()) == 3
        return Output(out=a_squared_plus_ab, res=b_plus_2a)

    outputs = await client.run_graph(foo, a=2, b=3)
    assert outputs == {"out": IntValue(10), "res": IntValue(7)}


@pytest.mark.asyncio
async def test_unpacking(
    bi,
    client: RuntimeClient,
    sig: Signature,
) -> None:
    def num_unpacks() -> int:
        return sum(1 for n in current_graph().nodes() if n.is_unpack_node())

    pn = bi["python_nodes"]

    @graph()
    def foo(a: Input[FloatValue], b: Input[IntValue]) -> Output:
        f = bi.make_struct(foo=a, bar=b)
        id_: "NodeRef" = pn.id_py(value=f["struct"])
        sturct: "NodePort" = id_["value"]
        old_proto = current_graph().to_proto()

        foo: "StablePortFunc" = sturct["foo"]
        assert num_unpacks() == 0
        assert current_graph().to_proto() == old_proto  # Nothing changed yet

        first = bi.id(foo)
        assert num_unpacks() == 1
        second = bi.iadd(a=sturct["bar"], b=Const(1))
        assert num_unpacks() == 1

        # Repeated uses of the same struct member need an explicit copy
        proto = current_graph().to_proto()
        with pytest.raises(ValueError, match="Already an edge from"):
            current_graph().add_edge(
                sturct["foo"].resolve(), current_graph().output["second"]
            )
        assert proto == current_graph().to_proto()  # Did nothing
        return Output(first=first, second=second)

    outputs = await client.run_graph(foo, a=3.142, b=5)
    assert outputs == {"first": FloatValue(3.142), "second": IntValue(6)}


def num_copies_unpacks() -> Tuple[int, int]:
    return (
        sum(1 for n in current_graph().nodes() if n.is_copy_node()),
        sum(1 for n in current_graph().nodes() if n.is_unpack_node()),
    )


def test_cant_unpack_original_after_copy(bi) -> None:
    @graph()
    def foo(a: Input[StringValue], b: Input[FloatValue]) -> Output:
        sturct: NodePort = bi.make_struct(foo=a, bar=b)["struct"]

        bi.discard(sturct["foo"])
        assert num_copies_unpacks() == (0, 1)

        cp = Copyable(sturct)
        o: Output = Output(whole=cp)
        assert num_copies_unpacks() == (1, 1)

        proto = current_graph().to_proto()
        with pytest.raises(ValueError, match="Cannot unpack"):
            current_graph().add_edge(
                sturct["bar"].resolve(), current_graph().output["second"]
            )
        assert proto == current_graph().to_proto()  # Did nothing
        return o


@pytest.mark.asyncio
async def test_can_unpack_copy_with_resolve(bi, client: RuntimeClient) -> None:
    @graph()
    def foo(a: Input[IntValue], b: Input[IntValue]) -> Output:
        sturct = bi.make_struct(foo=a, bar=b)["struct"]

        s = bi.iadd(a=sturct["foo"], b=sturct["bar"])
        assert num_copies_unpacks() == (0, 1)

        cp = Copyable(sturct).resolve()

        o: Output = Output(sum=s, product=bi.imul(a=cp["foo"], b=cp["bar"]))
        assert num_copies_unpacks() == (1, 2)
        return o

    outputs = await client.run_graph(foo, a=3, b=4)
    assert outputs == {"sum": IntValue(7), "product": IntValue(12)}


@pytest.mark.asyncio
async def test_Copyable_fields(bi, client: RuntimeClient) -> None:
    @graph()
    def foo(a: Input[IntValue], b: Input[IntValue]) -> Output:
        sturct = bi.make_struct(foo=a, bar=b)["struct"]

        s = bi.iadd(a=sturct["foo"], b=sturct["bar"])
        assert num_copies_unpacks() == (0, 1)

        return Output(
            sum=s,
            product=bi.imul(a=Copyable(sturct["foo"]), b=Copyable(sturct["bar"])),
        )

    outputs = await client.run_graph(foo, a=3, b=4)
    assert outputs == {"sum": IntValue(7), "product": IntValue(12)}


@pytest.mark.asyncio
async def test_unpacking_nested(client: RuntimeClient) -> None:
    @graph()
    def foo(arg: Input[Any]) -> Output:
        return Output(out=arg["outer"]["inner"])

    nested_val = FloatValue(42.0)
    argument: StructValue = StructValue(
        values={"outer": StructValue(values={"inner": nested_val})}
    )

    outputs = await client.run_graph(foo, arg=argument)
    assert outputs == {"out": nested_val}


def test_retry_secs(bi) -> None:
    pn = bi["python_nodes"]

    @graph()
    def foo(arg: Input[Any]) -> Output:
        return Output(res=pn.id_py(arg).with_retry_secs(2))

    (f,) = [n for n in foo.nodes() if isinstance(n, FunctionNode)]
    assert f.retry_secs == 2


@pytest.mark.asyncio
async def test_bad_annotations() -> None:
    # pyright does not support no_type_check decorator
    # so peppered with type: ignore
    with pytest.raises(TypeError, match="return type"):

        @no_type_check
        @graph()  # type: ignore
        def foo1(arg: Input[Any]) -> int:
            return Output(out=arg)  # type: ignore

    with pytest.raises(TypeError, match="return type"):

        @no_type_check
        @graph()
        def foo2(arg: Input[Any]) -> Output[int]:  # type: ignore
            return Output(out=arg)

    with pytest.raises(TypeError, match="Graph builder function arguments"):

        @no_type_check
        @lazy_graph()
        def foo3(arg: float) -> Output:
            return Output(out=arg)  # type: ignore

    with pytest.raises(ValueError, match="Cannot convert"):

        @no_type_check
        @lazy_graph()
        def foo4(arg: Input[float]) -> Output:  # type: ignore
            return Output(out=arg)


@pytest.mark.skip_typecheck
@pytest.mark.asyncio
async def test_box_order(bi: Namespace, sig: Signature) -> None:
    @dataclass
    class TestOut(UnpackRow):
        first: IntValue
        lst: VecValue[PairValue[IntValue, StringValue]]

    @graph(type_check_sig=sig)
    def push_pair(lst, first, second) -> Output[TestOut]:
        first = Copyable(first)
        pair = bi.make_pair(first, second)
        return Output(lst=bi.push(lst, pair), first=first)

    assert push_pair.input_order == ["lst", "first", "second"]
    assert push_pair.output_order == ["first", "lst"]

    @graph(type_check_sig=sig)
    def test_g() -> Output:
        i, lst = push_pair(Const([]), Const(3), Const("hello"))
        return Output(one=lst, two=i)

    box_n_name = next(
        n for n, nod in enumerate(test_g.nodes()) if isinstance(nod, BoxNode)
    )
    box_ins = {
        p2: n1
        for n1, _, (_, p2) in test_g._graph.in_edges(box_n_name, keys=True)  # type: ignore
    }
    assert len(box_ins) == 3

    assert all(isinstance(test_g[n], ConstNode) for n in box_ins.values())
    assert isinstance(cast(ConstNode, test_g[box_ins["lst"]]).value, VecValue)
    assert isinstance(cast(ConstNode, test_g[box_ins["first"]]).value, IntValue)
    assert isinstance(cast(ConstNode, test_g[box_ins["second"]]).value, StringValue)

    box_outs = tuple(
        ports
        for _, _, ports in test_g._graph.out_edges(box_n_name, keys=True)  # type: ignore
    )
    assert len(box_outs) == 2

    assert ("first", "two") in box_outs
    assert ("lst", "one") in box_outs


@pytest.mark.asyncio
async def test_scope(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        with Scope():
            bi.discard(Const(3))
        assert len(current_builder().inner_scopes.values()) == 1
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 3
    n = cast(
        BoxNode,
        next(filter(lambda x: isinstance(x, BoxNode), tc_graph.nodes())),
    )
    assert any([x.is_discard_node() for x in n.graph.nodes()])
    assert n.graph.inputs() == []
    assert n.graph.outputs() == []


@pytest.mark.asyncio
async def test_scope_capture_in(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Const(3)
        with Scope():
            bi.discard(a)
        assert len(current_builder().inner_scopes.values()) == 1
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 4
    n = cast(
        BoxNode,
        next(filter(lambda x: isinstance(x, BoxNode), tc_graph.nodes())),
    )
    assert any([x.is_discard_node() for x in n.graph.nodes()])
    assert n.graph.inputs() == ["_c0"]
    assert n.graph.outputs() == []


@pytest.mark.asyncio
async def test_scope_capture_out(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        with Scope():
            a = Const(3)
        assert len(current_builder().inner_scopes.values()) == 1
        bi.discard(a)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 4
    n = cast(
        BoxNode,
        next(filter(lambda x: isinstance(x, BoxNode), tc_graph.nodes())),
    )
    assert not any([x.is_discard_node() for x in n.graph.nodes()])
    assert n.graph.inputs() == []
    assert n.graph.outputs() == ["_c0"]


@pytest.mark.asyncio
async def test_nested_scopes(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Const(3)
        with Scope():
            b = Const(2)
            c = bi.iadd(a, b)
            with Scope():
                d, e = bi.copy(c)
            assert len(current_builder().inner_scopes.values()) == 1
            with Scope():
                f = bi.iadd(d, Const(1))
                assert len(current_builder().inner_scopes.values()) == 0
            assert len(current_builder().inner_scopes.values()) == 2
        assert len(current_builder().inner_scopes.values()) == 3
        assert e.node_ref.graph in current_builder().inner_scopes
        h = bi.iadd(e, f)

        return Output(value=h)

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 5
    n = cast(
        BoxNode,
        next(filter(lambda x: isinstance(x, BoxNode), tc_graph.nodes())),
    )
    assert any([isinstance(x, BoxNode) for x in n.graph.nodes()])
    assert n.graph.inputs() == ["_c0"]
    assert sorted(n.graph.outputs()) == ["_c0", "_c1"]


@pytest.mark.asyncio
async def test_copyable_capture_in(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Copyable(Const(3))
        with Scope():
            bi.discard(a)
        with Scope():
            bi.discard(a)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 6
    assert any([n.is_copy_node() for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_copyable_capture_out_once(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        with Scope():
            a = Copyable(Const(3))
        bi.discard(a)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 4
    assert not any([n.is_copy_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_copyable_capture_out_copied(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        with Scope():
            a = Copyable(Const(3))
            bi.discard(a)
        bi.discard(a)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 4
    assert not any([n.is_copy_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_copyable_on_captured_input(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Const(3)
        with Scope():
            b = Copyable(a)
            bi.discard(b)
        bi.discard(b)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 5
    assert not any([n.is_copy_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_copyable_dont_use(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Const(3)
        with Scope():
            b = Copyable(a)
        bi.discard(b)
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 5
    assert not any([n.is_copy_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_unpack_capture_in(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        a = Const(StructValue({"a": IntValue(1), "b": StringValue("hello")}))
        with Scope():
            bi.discard(a["a"])
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 6
    assert any([n.is_unpack_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.asyncio
async def test_unpack_capture_out(bi, client: RuntimeClient) -> None:
    @graph()
    def g() -> Output:
        with Scope():
            a = Const(StructValue({"a": IntValue(1), "b": StringValue("hello")}))
        bi.discard(a["a"])
        return Output()

    if not client.can_type_check:
        pytest.skip()
    tc_graph = await client.type_check_graph(g)
    assert tc_graph.n_nodes == 4
    assert not any([n.is_unpack_node() for n in tc_graph.nodes()])
    assert any([isinstance(n, BoxNode) for n in tc_graph.nodes()])


@pytest.mark.skip_typecheck
@pytest.mark.asyncio
async def test_parmap_builder(bi: Namespace, sig, pyruntime: PyRuntime) -> None:
    client = pyruntime  # We expect to make this pass for Rust runtime too soon

    @lazy_graph()
    def const_vec() -> Output:
        return Output(vec=Const([]))

    @lazy_graph(type_check_sig=sig)
    def par_map_builder(f, ins) -> Output:
        """Returns a graph which when evaluated applies f to each value in inps
        and collates the results."""
        f = bi.parallel(f, Const(rename_ports_graph({"vec": "vec"})))
        f = bi.sequence(f, Const(bi.push.to_graph(input_map={"item": "out"})))

        @loop()
        def built(loop_dat) -> Output:
            vec: ValueSource = Copyable(loop_dat["vec"])
            with IfElse(bi.eq(Const([]), vec)) as loop_ifelse:
                with If():
                    Break(loop_dat["big_g"])
                with Else():
                    vec, i = bi.pop(vec)

                    f_part = bi.partial(f, inp=i)

                    big_g = bi.sequence(loop_dat["big_g"], f_part)
                    Continue(bi.make_struct(vec=vec, big_g=big_g))
            return Output(loop_ifelse.nref)

        par_f = built(bi.make_struct(vec=ins, big_g=Const(const_vec)))

        return Output(par_f)

    @graph()
    def func(inp) -> Output:
        return Output(out=bi.imul(inp, Const(2)))

    ins = [2387, 123, 64]
    out = await client.run_graph(par_map_builder.graph, f=func, ins=ins)

    gv = out["value"]

    assert isinstance(gv, GraphValue)

    tg = gv.value

    @graph()
    def expected_graph() -> Output:
        ls: ValueSource = Const([])
        for v in reversed(ins):
            ls = bi.push(ls, bi.imul(Const(v), Const(2)))
        return Output(vec=ls)

    # tg contains some type info, but has never been fully type-checked.
    # expected_graph contains no type information.
    # For comparison, we must make the two contain the same type information.
    tg = await client.type_check_graph(tg)
    exp = await client.type_check_graph(expected_graph)
    tg.name = exp.name
    assert _compare_graphs(tg, exp)
    out = await client.run_graph(tg)

    assert out["vec"].try_autopython() == [x * 2 for x in reversed(ins)]



import inspect
from dataclasses import dataclass, fields, is_dataclass
from typing import (
    Any,
    ParamSpec,
    Type,
    cast,
    get_origin,
    get_type_hints,
)

import pydantic as pyd

from tierkreis.core.opaque_model import OpaqueModel

Out = tuple[Type, str | None]


@dataclass(frozen=True)
class FieldExtractionError(Exception):
    """An error occurred while extracting fields from a class."""

    type_: Type | ParamSpec
    msg: str


def _assert_annotation(annotation: Type | None, struct_type: Type) -> Type:
    if annotation is None:
        raise FieldExtractionError(
            struct_type, "Classes without type annotations cannot be converted."
        )

    return annotation


@dataclass(frozen=True)
class ClassField:
    """Capture required data about a class field."""

    name: str
    type_: Type
    # the discriminant tag for the field, if it is a discriminated union
    discriminant: str | None
    # whether the field is present in the __init__ function
    init: bool = True
    default: Any = None


def python_struct_fields(
    type_: Type | ParamSpec,
) -> list[ClassField]:
    """For a python dataclass or pydantic BaseModel, extract the fields and their types."""
    if inspect.isclass(type_) and issubclass(type_, pyd.BaseModel):
        if issubclass(type_, OpaqueModel):
            model_type = cast(Type[OpaqueModel], type_)

            return [
                ClassField(
                    model_type.tierkreis_field(),
                    str,
                    None,
                )
            ]
        model_type = cast(Type[pyd.BaseModel], type_)
        model_fields = model_type.model_fields

        def _get_discriminator(f: pyd.fields.FieldInfo) -> str | None:
            disc = f.discriminator
            if isinstance(disc, pyd.Discriminator):
                raise ValueError("Discriminators must be static strings.")
            return disc

        return [
            ClassField(
                k,
                _assert_annotation(f.annotation, type_),
                _get_discriminator(f),
                default=f.default,
            )
            for k, f in model_fields.items()
        ]
    # pydantic binds concrete types to generic fields when available in the
    # annotation.
    # For generic dataclasses, just deal with the generic base type (used in workers).
    type_ = get_origin(type_) or type_
    if is_dataclass(type_):
        dat_fields = fields(type_)
        types = get_type_hints(type_)
        return [
            ClassField(
                f.name,
                types[f.name],
                getattr(f.default, "discriminator", None),
                default=f.default,
            )
            for f in dat_fields
        ]
    raise FieldExtractionError(
        type_, "Can only convert dataclasses or pydantic BaseModel."
    )


def generic_origin(type_: Type) -> Type | None:
    """Like typing.get_origin but also supports Generic pydantic 'BaseModel's"""
    if (o := get_origin(type_)) is not None:
        return o
    if inspect.isclass(type_) and issubclass(type_, pyd.BaseModel):
        # This taken from
        # https://github.com/pydantic/pydantic/blob/812516d71a8696d5e29c5bdab40336d82ccde412/pydantic/_internal/_generics.py#L214-L218
        pydantic_generic_metadata = getattr(
            type_, "__pydantic_generic_metadata__", None
        )
        if pydantic_generic_metadata:
            return pydantic_generic_metadata.get("origin")
        # BaseModel but not generic; fall through
    return None



"""Worker exceptions."""

from dataclasses import dataclass


class FunctionNotFound(Exception):
    """Function not found."""

    def __init__(self, function):
        super().__init__()
        self.function = function


class DecodeInputError(Exception):
    pass


class EncodeOutputError(Exception):
    pass


@dataclass
class NodeExecutionError(Exception):
    base_exception: Exception


@dataclass(frozen=True)
class NamespaceClash(Exception):
    namespace: list[str]
    functions: set[str]

    def __str__(self) -> str:
        return f"""Clash in namespace {'::'.join(self.namespace)} of functions\n
        {self.functions}"""



from importlib.util import find_spec
from typing import Optional, Tuple, Union, overload

import betterproto

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps

# Awkwardly, the Rust stubs end up here:
from tierkreis import TierkreisGraph
from tierkreis.core.signature import Namespace, Signature
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.values import StructValue

try:
    _TYPE_CHECK = find_spec("tierkreis_typecheck") is not None
except ModuleNotFoundError:
    _TYPE_CHECK = False


class TypeCheckNotInstalled(ImportError):
    pass


_ERR = TypeCheckNotInstalled(
    "Type checking requires tierkreis_typecheck package to be installed."
)


@overload
def infer_graph_types(
    g: TierkreisGraph,
    funcs: Signature,
    inputs: None = None,
) -> TierkreisGraph: ...


@overload
def infer_graph_types(
    g: TierkreisGraph,
    funcs: Signature,
    inputs: StructValue,
) -> Tuple[TierkreisGraph, StructValue]: ...


def infer_graph_types(
    g: TierkreisGraph,
    funcs: Signature,
    inputs: Optional[StructValue] = None,
) -> Union[TierkreisGraph, Tuple[TierkreisGraph, StructValue]]:
    if not _TYPE_CHECK:
        raise _ERR
    else:
        import tierkreis_typecheck

    req = ps.InferGraphTypesRequest(
        gwi=ps.GraphWithInputs(
            graph=g.to_proto(),
            inputs=None
            if inputs is None
            else pg.StructValue(map=inputs.to_proto_dict()),
        ),
        functions=funcs.root.to_proto(),
    )
    resp = ps.InferGraphTypesResponse().parse(
        tierkreis_typecheck.infer_graph_types(bytes(req))
    )
    name, _ = betterproto.which_one_of(resp, "response")
    if name == "success":
        g = TierkreisGraph.from_proto(resp.success.graph)
        if inputs is None:
            assert resp.success.inputs is None
            return g
        return (g, StructValue.from_proto(resp.success.inputs))
    raise TierkreisTypeErrors.from_proto(resp.error, g)


def builtin_namespace() -> Namespace:
    if _TYPE_CHECK:
        import tierkreis_typecheck

        return Namespace.from_proto(
            ps.Namespace().parse(tierkreis_typecheck.builtin_namespace())
        )
    else:
        raise _ERR



import asyncio
import re
import sys
import traceback
from contextlib import asynccontextmanager
from functools import wraps
from pathlib import Path
from signal import SIGINT, SIGTERM
from typing import TYPE_CHECKING, AsyncContextManager, Dict, List, Optional, cast

import click
from grpclib.client import Channel
from yachalk import chalk

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis import TierkreisGraph
from tierkreis.builder import _func_sig
from tierkreis.client import ServerRuntime
from tierkreis.core.graphviz import tierkreis_to_graphviz
from tierkreis.core.protos.tierkreis.v1alpha1.graph import Graph as ProtoGraph
from tierkreis.core.signature import Signature
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.types import StructType
from tierkreis.core.values import StructValue, TierkreisValue

if TYPE_CHECKING:
    from tierkreis.core.function import FunctionDeclaration


@asynccontextmanager
async def server_manager(host: str, port: int = 443):
    async with Channel(host, port) as channel:
        yield ServerRuntime(channel)


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def _inputs(proto_file: str, py_source: str) -> Dict[str, TierkreisValue]:
    if proto_file == "":
        return (
            {}
            if py_source == ""
            else {k: TierkreisValue.from_python(v) for k, v in eval(py_source).items()}
        )
    if py_source != "":
        raise ValueError("Cannot provide both inputs and --py-inputs")
    path = Path(proto_file)
    with open(path, "rb") as f:
        v: StructValue = StructValue.from_proto(pg.StructValue().parse(f.read()))
        return v.values


async def _parse(source: Path) -> TierkreisGraph:
    with open(source, "rb") as f:
        return TierkreisGraph.from_proto(ProtoGraph().parse(f.read()))


async def _check_graph(
    source_path: Path,
    client_manager: AsyncContextManager[ServerRuntime],
) -> TierkreisGraph:
    async with client_manager as client:
        tkg = await _parse(source_path)
        try:
            tkg = await client.type_check_graph(tkg)
        except TierkreisTypeErrors:
            _print_typeerrs(traceback.format_exc(0))
            sys.exit(1)
        return tkg


async def main_coro(manager: AsyncContextManager):
    async with manager as runtime:
        try:
            print(runtime.socket_address(), flush=True)
            await asyncio.sleep(10000000000000)
        except asyncio.CancelledError:
            print("\nShutting Down")


def run_with_signals(manager: AsyncContextManager):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    main_task = loop.create_task(main_coro(manager))
    for signal in [SIGINT, SIGTERM]:
        loop.add_signal_handler(signal, main_task.cancel)
    try:
        loop.run_until_complete(main_task)
    finally:
        loop.close()


@click.group()
@click.pass_context
@click.option(
    "--runtime",
    "-r",
    default="localhost",
    help="Choose runtime, default=localhost",
)
@click.option(
    "--port",
    "-p",
    help="Runtime port, default=8090 if runtime is localhost, else 443",
)
@coro
async def cli(ctx: click.Context, runtime: str, port: Optional[int]):
    local = runtime == "localhost"
    if port is None:
        port = 8090 if local else 443
    ctx.ensure_object(dict)
    ctx.obj["runtime_label"] = runtime
    client_manager = server_manager(runtime, port)
    asyncio.get_event_loop()
    ctx.obj["client_manager"] = client_manager


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.pass_context
@coro
async def check(ctx: click.Context, proto: str) -> TierkreisGraph:
    """Type check PROTO binary file against runtime signature."""
    source_path = Path(proto)
    tkg = await _check_graph(source_path, ctx.obj["client_manager"])
    print(chalk.bold.green("Success: graph type check complete."))
    return tkg


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.argument(
    "view_path",
    type=click.Path(exists=False),
)
@click.option("--check", "-C", is_flag=True, help="Type check and annotate graph.")
@click.option(
    "--unbox-level",
    default=0,
    help="Nesting level to which boxes and thunks should be unthunked, default=0.",
)
@click.pass_context
@coro
async def view(
    ctx: click.Context,
    proto: str,
    view_path: str,
    check: bool,
    unbox_level: int,
):
    """Visualise PROTO binary as graph and output to VIEW_PATH."""
    source_path = Path(proto)
    if check:
        tkg = await _check_graph(
            source_path,
            ctx.obj["client_manager"],
        )
    else:
        tkg = await _parse(source_path)

    tkg.name = source_path.stem
    view_p = Path(view_path)
    ext = view_p.suffix
    tierkreis_to_graphviz(tkg, unbox_level=unbox_level).render(
        view_path[: -len(ext)], format=ext[1:]
    )


def _print_outputs(outputs: dict[str, TierkreisValue]):
    print(
        "\n".join(
            f"{chalk.bold.yellow(key)}: {val.viz_str()}" for key, val in outputs.items()
        )
    )


def _print_typeerrs(errs: str):
    print(chalk.red(errs), file=sys.stderr)


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.argument("inputs", default="")
@click.option("-p", "--py-inputs", default="")
@click.pass_context
@coro
async def run(ctx: click.Context, proto: Path, inputs: str, py_inputs: str):
    """Run PROTO binary on runtime with optional INPUTS and output to console."""
    async with ctx.obj["client_manager"] as client:
        client = cast(ServerRuntime, client)
        tkg = await _parse(proto)
        input_dict = _inputs(inputs, py_inputs)
        try:
            outputs = await client.run_graph(tkg, **input_dict)
            _print_outputs(outputs)
        except TierkreisTypeErrors:
            _print_typeerrs(traceback.format_exc(0))
            sys.exit(1)


PORT_RE = re.compile(r"([\w]+):")


def _print_namespace(sig: Signature, namespace: List[str], function: Optional[str]):
    namespace_str = "::".join(namespace)
    root = "_root"

    print(chalk.bold(f"Namespace: {root if namespace == [] else namespace_str}"))
    print()
    print(chalk.bold("Aliases and Struct definitions"))

    for alias, type_scheme in sig.aliases.items():
        prefix = f"{namespace_str}::"
        if not alias.startswith(prefix):
            continue
        alias = alias[len(prefix) :]
        type_ = type_scheme.body
        if isinstance(type_, StructType):
            alias_string = type_.anon_name()
            alias_string = PORT_RE.sub(
                lambda match: chalk.yellow(match.group()), alias_string
            )
        else:
            alias_string = str(type_)
        print(f"{chalk.bold.magenta(alias)} = {alias_string}\n")

    print()
    print(chalk.bold("Functions"))

    names_dict = sig.root.get(namespace).functions
    func_names = [function] if function else list(names_dict.keys())
    for name in sorted(func_names):
        _print_func(name, names_dict[name])
        print()


def _print_func(name: str, func: "FunctionDeclaration"):
    print(_func_sig(name, func))
    if func.description:
        print(chalk.green(func.description))


@cli.command()
@click.pass_context
@click.option("--namespace", type=str, help="Show only signatures of this namespace.")
@click.option(
    "--function", type=str, help="Show only the signature of a particular function."
)
@coro
async def signature(
    ctx: click.Context, namespace: Optional[str], function: Optional[str]
):
    """Check signature of available namespaces and functions on runtime."""
    async with ctx.obj["client_manager"] as client:
        client = cast(ServerRuntime, client)
        label = ctx.obj["runtime_label"]
        print(chalk.bold(f"Runtime: {label}"))
        print()
        sig = await client.get_signature()
        namespaces: list[list[str]] = (
            [namespace.split("::")] if namespace else sig.root.all_namespaces()
        )

        for ns in namespaces:
            _print_namespace(sig, ns, function)
            print()



class Labels:
    def __init__(self):
        raise RuntimeError("Do not instantiate")

    THUNK = "thunk"
    VALUE = "value"
    VARIANT_VALUE = "variant_value"
    CONTINUE = "continue"
    BREAK = "break"


# This needs to be here to fix some cycles in the imports
from .tierkreis_graph import TierkreisGraph  # noqa: E402, F401



from dataclasses import dataclass
from time import time
from typing import Any, Dict, List, Optional, Tuple, Type

import pytest
from utils import nint_adder

from tierkreis import TierkreisGraph
from tierkreis.client import RuntimeClient
from tierkreis.core.graphviz import _merge_copies
from tierkreis.core.tierkreis_graph import FunctionNode
from tierkreis.core.utils import graph_from_func
from tierkreis.core.values import FloatValue, VariantValue
from tierkreis.pyruntime import PyRuntime
from tierkreis.worker.exceptions import NodeExecutionError


@pytest.mark.asyncio
async def test_nint_adder(client: RuntimeClient):
    for in_list in ([1] * 5, list(range(5))):
        tk_g = nint_adder(len(in_list))
        outputs = await client.run_graph(tk_g, array=in_list)
        assert outputs["out"].try_autopython() == sum(in_list)


def add_n_graph(increment: int) -> TierkreisGraph:
    tk_g = TierkreisGraph()

    add_func = tk_g.add_func(
        "iadd", a=tk_g.add_const(increment), b=tk_g.input["number"]
    )
    tk_g.set_outputs(output=add_func)

    return tk_g


@pytest.mark.asyncio
async def test_switch(client: RuntimeClient):
    add_2_g = add_n_graph(2)
    add_3_g = add_n_graph(3)
    tk_g = TierkreisGraph()

    switch = tk_g.add_func(
        "switch",
        if_true=tk_g.add_const(add_2_g),
        if_false=tk_g.add_const(add_3_g),
        pred=tk_g.input["flag"],
    )

    eval_node = tk_g.add_func("eval", thunk=switch, number=tk_g.input["number"])

    tk_g.set_outputs(out=eval_node["output"])

    outs = await client.run_graph(tk_g, flag=True, number=3)

    assert outs["out"].try_autopython() == 5
    outs = await client.run_graph(tk_g, flag=False, number=3)

    assert outs["out"].try_autopython() == 6


@pytest.mark.asyncio
async def test_match(client: RuntimeClient):
    # Test a variant type < one: Float | many: Vec<Float> >
    one_graph = TierkreisGraph()
    one_graph.set_outputs(
        value=one_graph.add_func(
            "fadd", a=one_graph.input["value"], b=one_graph.add_const(3.14)
        )
    )
    many_graph = TierkreisGraph()
    many_graph.set_outputs(
        value=many_graph.add_func("pop", vec=many_graph.input["value"])["item"]
    )

    match_graph = TierkreisGraph()
    match_graph.set_outputs(
        result=match_graph.add_func(
            "python_nodes::id_delay",
            wait=match_graph.add_const(1),
            value=match_graph.add_func(
                "eval",
                thunk=match_graph.add_match(
                    match_graph.input["vv"],
                    one=match_graph.add_const(one_graph),
                    many=match_graph.add_func(
                        "python_nodes::id_delay",
                        wait=match_graph.add_const(1),
                        value=match_graph.add_const(many_graph),
                    ),
                )["thunk"],
            ),
        )
    )

    start_time = time()
    outs = await client.run_graph(match_graph, vv=VariantValue("one", FloatValue(6.0)))
    time_taken = time() - start_time
    assert outs["result"] == FloatValue(9.14)
    # Must have waited at least 1s for the delay on the graph output.
    assert time_taken >= 1.0

    if isinstance(client, PyRuntime):
        # PyRuntime waits for all inputs to a node to be ready.
        assert time_taken > 2.0
    else:
        # Should not have had to wait 1s first for the "many" graph input to be ready,
        # as that was not the variant was not selected.
        # (1.5s is arbitrary, but less than 2s.)
        assert time_taken < 1.5


@pytest.mark.asyncio
async def test_tag(client: RuntimeClient):
    g = TierkreisGraph()
    g.set_outputs(res=g.add_tag("foo", value=g.input["arg"]))
    v = FloatValue(67.1)
    outs = await client.run_graph(g, arg=v)
    assert outs == {"res": VariantValue("foo", v)}


@dataclass
class NestedStruct:
    s: List[int]
    a: Tuple[int, bool]
    b: Optional[str]
    d: Optional[float]


@dataclass
class TstStruct:
    x: int
    y: bool
    m: Dict[int, int]
    n: NestedStruct


@pytest.mark.asyncio
async def test_idpy(client: RuntimeClient, idpy_graph: TierkreisGraph):
    async def assert_id_py(val: Any, typ: Type) -> bool:
        output = await client.run_graph(idpy_graph, id_in=val)
        val_decoded = output["id_out"].to_python(typ)
        return val_decoded == val

    dic: Dict[int, bool] = {1: True, 2: False}

    nestst = NestedStruct([1, 2, 3], (5, True), "asdf", None)
    testst = TstStruct(2, False, {66: 77}, nestst)
    pairs: list[tuple[Any, Type]] = [
        (dic, dict[int, bool]),
        (testst, TstStruct),
        ("test123", str),
        (2, int),
        (132.3, float),
        ((2, "a"), tuple[int, str]),
        ([1, 2, 3], list[int]),
        (True, bool),
    ]
    for val, typ in pairs:
        assert await assert_id_py(val, typ)


@pytest.mark.asyncio
async def test_fail_node(client: RuntimeClient) -> None:
    tg = TierkreisGraph()
    tg.add_func("python_nodes::fail", value=tg.input["arg"])

    exception: Type[Exception] = (
        NodeExecutionError if isinstance(client, PyRuntime) else RuntimeError
    )
    with pytest.raises(exception) as err:
        await client.run_graph(tg, arg=3)

    assert "fail node was run" in str(err.value)


@pytest.mark.asyncio
async def test_vec_sequence(client: RuntimeClient) -> None:
    sig = await client.get_signature()
    pop_g = graph_from_func("pop", sig.root.functions["pop"])
    push_g = graph_from_func("push", sig.root.functions["push"])

    seq_g = graph_from_func("sequence", sig.root.functions["sequence"])

    outputs = await client.run_graph(seq_g, first=pop_g, second=push_g)

    sequenced_g: TierkreisGraph = (
        outputs["sequenced"].to_python(TierkreisGraph).inline_boxes()
    )

    # check composition is succesful
    functions = {
        node.function_name.name
        for node in sequenced_g.nodes()
        if isinstance(node, FunctionNode)
    }
    assert {"push", "pop"}.issubset(functions)

    # check composition evaluates
    vec_in = ["foo", "bar", "bang"]
    out = await client.run_graph(sequenced_g, vec=vec_in)
    vec_out = out["vec"].to_python(List[str])
    assert vec_in == vec_out


def test_merge_copies():
    tg = TierkreisGraph()
    x1, x2 = tg.copy_value(tg.input["x"])
    x3, x4 = tg.copy_value(x1)
    y1, y2 = tg.copy_value(x3)
    z1, z2 = tg.copy_value(x4)

    tg.set_outputs(x2=x2, y1=y1, y2=y2, z1=z1, z2=z2)

    assert tg.n_nodes == 6
    assert sum(1 for _ in tg.edges()) == 9

    tg = _merge_copies(tg)

    assert tg.n_nodes == 3
    assert sum(1 for _ in tg.edges()) == 6


@pytest.mark.asyncio
async def test_subspaces(client: RuntimeClient):
    tg = TierkreisGraph()
    idnode = tg.add_func("python_nodes::subspace::increment", value=tg.add_const(0))
    tg.set_outputs(out=idnode)

    assert (await client.run_graph(tg))["out"].try_autopython() == 2

    tg = TierkreisGraph()
    idnode = tg.add_func("python_nodes::increment", value=tg.add_const(0))
    tg.set_outputs(out=idnode)

    assert (await client.run_graph(tg))["out"].try_autopython() == 1


@pytest.mark.asyncio
async def test_map(client: RuntimeClient):
    thunk = TierkreisGraph()
    double = thunk.add_func("imul", a=thunk.input["value"], b=thunk.add_const(2))
    thunk.set_outputs(value=double)

    tg = TierkreisGraph()
    map_node = tg.add_func(
        "map", value=tg.add_const([1, 2, 3]), thunk=tg.add_const(thunk)
    )
    tg.set_outputs(value=map_node)

    assert (await client.run_graph(tg))["value"].try_autopython() == [2, 4, 6]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/runtime.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    graph as _graph__,
    signature as _signature__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Callback(betterproto.Message):
    uri: str = betterproto.string_field(1)
    loc: "_graph__.Location" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class RunGraphRequest(betterproto.Message):
    graph: "_graph__.Graph" = betterproto.message_field(1)
    inputs: "_graph__.StructValue" = betterproto.message_field(2)
    type_check: bool = betterproto.bool_field(3)
    loc: "_graph__.Location" = betterproto.message_field(4)
    escape: "Callback" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class RunGraphResponse(betterproto.Message):
    success: "_graph__.StructValue" = betterproto.message_field(1, group="result")
    error: str = betterproto.string_field(2, group="result")
    type_errors: "_signature__.TypeErrors" = betterproto.message_field(
        3, group="result"
    )


class RuntimeStub(betterproto.ServiceStub):
    async def run_graph(
        self,
        run_graph_request: "RunGraphRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RunGraphResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.runtime.Runtime/RunGraph",
            run_graph_request,
            RunGraphResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class RuntimeBase(ServiceBase):

    async def run_graph(
        self, run_graph_request: "RunGraphRequest"
    ) -> "RunGraphResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_run_graph(
        self, stream: "grpclib.server.Stream[RunGraphRequest, RunGraphResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.run_graph(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.runtime.Runtime/RunGraph": grpclib.const.Handler(
                self.__rpc_run_graph,
                grpclib.const.Cardinality.UNARY_UNARY,
                RunGraphRequest,
                RunGraphResponse,
            ),
        }
